# TReqS concepts

## Entities

TReqS is based on 5 entities. 

* TreqsRequest, TreqsFile and TreqsTape are dynamic entities created when a staging request is received. 

* TreqsUser and TreqsTapeModel are static entities created by administrative task (potentially automatically at startup)

* When a staging request is issued, a TreqsRequest is always created.  A TreqsRequest is related to one, and only one TreqsFile. 
Several requests can be created for the same TreqsFile.

* A TreqsFile is created for each distinct file provided through the request, whether this file exists or not into HPSS.
Several TreqsRequest, created by different users or by the same one, and addressing the same file, will lead to only one 
creation of a TreqsFile object.

* A TreqsTape represents a tape and a list of TreqsFile located on the same HPSS tape. 

* A TreqsUser represents a user with its authentication and authorization attributes. 

* A TreqsTapeModel represents the model of a tape.

## Dispatching

When several TreqsFile belong to the same tape, they are grouped into the same queue (ie a TreqsTape that is not processed yet).
To improve staging throughput, files are dynamically sorted according to the HPSS metadata named 'position on tape' 
which represents the logical file position on the tape.

Moreover, TReqS is able to insert new TreqsFile into a queue even if the staging process has already been started for that queue.

There is one active queue for each tape. The number of active queues is not limited.

## Staging

The staging process consists of staging files that were previously aggregated per queue.
The staging is triggered at fixed time intervals and/or by admin command.
The number of simultaneous staging processes is limited per tape model.

The duration of the staging process involves several durations:

* The tape mount duration,
* The tape positioning duration, 
* The file reading duration (depending on the tape|drive reading rate),
* The network bandwidth,
* The disk writing rate.

A timeout is computed for each file staging based on a combination of the tape mount, 
the tape positioning and the tape reading rate. Network bandwidth and HPSS disk write rate are not
considered to have an impact on the total duration.

See "Managing a TReqS server" to know how to customize the staging process.

## TreqsRequest and TreqsFile statuses 

Statuses of those two entities are defined using the couple STATUS/SUB_STATUS. 
Here are available values:

* TreqsRequest:
  * SUBMITTED/- : The request is being processed.
  * ENDED/SUCCEEDED : The request has succeeded.
  * ENDED/FAILED : The request has failed.
  * ENDED/CANCELLED : The request has been cancelled.


* TreqsFile:
  * CREATED/- : The file has been taken into account, it may or may have not its HSM metadata yet.
  * DISPATCHED/- : THE HSM metadata have been retrieved, the file has been dispatched and is available for staging.
  * ENDED/ALREADYONDISK : The file will not be staged because it is already on HSM disk.  
  * STAGING/- : The file is currently being staged.
  * ENDED/STAGED : The file has been successfully staged.
  * ENDED/FAILED : There was an error when handling the file.
  * ENDED/CANCELLED : The staging of the file has been cancelled.

Statuses like 'ENDED/*' are considered as final states.

## Canceling requests

It is possible to cancel a staging request as soon as it is not in a final state. 
Canceling can be done per request, per file (all requests associated with that file) 
or per tape (all requests associated with all files located on that tape)

### Case 1: the file is still in queue (DISPATCHED)

The file associated with a cancelled request will be removed from the its staging queue only if no other current request
addresses it. For example, Bart and Lisa both request file '/path/to/springfield' at the same time. Then Bart cancels its
request. Bart's request will indeed be 'ENDED/CANCELLED' but the file '/path/to/springfield' will be staged anyway for Lisa.

### Case 2: the file/tape is currently being processed
  
* Canceling per request will have no effect on the staging process. 
The staging will not be aborted and will continue until the end.
The request will be marked as 'ENDED/CANCELLED', the file 'ENDED/STAGED'
 
* Canceling per file will be rejected with the message 
'The file is already being staged or in final state'.
 
* Canceling per tape while it is staging will have no staging effect on the file currently being staged, 
but next files in queue will be removed from the queue and marked as 'ENDED/CANCELLED'. 
Associated requests will be mark as 'ENDED/CANCELLED'.


## Archiving

In order to keep the TReqS H2 database small and efficient, ended 
requests (successfully or not) are not kept forever.
Once a request has reached a final state, it is archived (copied) in an external storage.
Have a look at the chapter "Managing a TReqS server" to know about different archiving solutions.
 
## Security model

### Authentication

Apart from 'administration/ping', all TReqS REST paths need the user to be authenticated with a password.
Password are stored (hashed) in the TReqS database.

### Authorization

Once a user is authenticated, it is authorized depending on its role 'ADMIN' or 'USER' or 'NONE'.

* A user with the role 'ADMIN' will be authorized to access all resources  (e.g. creating users or tapemodels, canceling any request)

* A user with the role 'USER' will be granted to (only) :
	* submit new requests 
  	* cancel its own requests only
  	* get information on resources : request, file, queue, tape, tapemodel
  	* get user' information for himself only (not all users)
  	* change its own password only
   
* A user with the role 'NONE' cannot access any resource. This is a convenient way to deny access to TReqS' resources.


## REST API

TReqS provides a REST API which allows you to interact with the server. 
Have a look at the documentation at 'doc/rest/overview-index.html' to have an better idea of how you can use TReqS as a
user or as an administrator.

If you are not comfortable with HTTP REST, you might want to use the Python client (based on the API).
See the README file for more information.


# Managing a TReqS server

## TReqS properties

Most of TReqS properties are customizable in the file 'etc/treqs2.properties'. Before you start TReqS, please have
a look at this file as comments might help you understand how to customize TReqS to fit your needs.

Dynamic properties are prefixed with 'treqs.dynamic.' and can be changed 'on the fly'. 
Static properties are NOT prefixed with 'treqs.dynamic.' and need a server restart.


## Managing Tape models

The TreqsTapeModel is used to control the number of allocated drives (e.g. 25 drives allocated for T10K-D model at CC-IN2P3).
We also use this entity to give a hint of reading rate in MiB/s per model. 
This value is used to compute, per file, a staging timeout taking into account the size of the file.

Since the model of a tape cannot be obtained from HPSS, At CC-IN2P3 tapes model's name is determined from tape's name.
For that we provide a configuration file 'etc/tapemodels.properties', mapping a regexp as tape name to a tape model.

There is no parsing order and the regexp expressing tape names are supposed to be exclusive.

Changes in this file require a server restart

A default tape model is specified using the 'default' keyword, like in the following excerpt :

```
 # KT0000 - KT1999 will be  'T10K-C'
 KT[0-1]([0-9]){3,}=T10K-C

 ## KT2000 - KTZZ99 will be  'T10K-D'
 KT[2-Z]([0-Z]){3,}=T10K-D

 # otherwise, default is T10K-D
 # default tapemodel will automatically be created at startup with 
 # 10 as max parallel staging and 100 MiB/s as reading rate
 default=T10K-D
```


In order to stage a file, the model of the tape associated with this file must exist, otherwise the request will fail.

Fortunately, at startup, TReqS creates a default tape model based on the 'default' property value in the 'etc/tapemodels.properties' with 
10 as max parallel staging and 100 MiB/s as reading rate.

However, in order to benefit TReqS's main features, it is recommended to define a set of models of tapes with appropriate values like 

For example, to create a new model named 'T10K-C' with a reading rate of 128MiB/s and allocate a maximum of 15 drives:

```
> curl -X POST -H 'Content-Type':'application/json' \
    -d '{"name": "T10K-C", "max_parallel_staging": "15", "reading_rate": "128"}' \
    http://treqs:changeit@localhost:8080/treqs2/administration/tapemodel 
```

You can also do this with the TReqS client.

## Managing users

TReqS comes with a default user 'treqs' with password 'changeit' and role ADMIN. 
It is strongly recommended to change this password.

To change the password of the default 'treqs' user to the new value 'NEWPASSWORD':

```
> curl -X PUT 'http://treqs:changeit@localhost:8080/treqs2/administration/authentication/user/treqs/password/NEWPASSORD'
```

To create a new user 'bart' with password 'mysecret':

```
> curl -X POST -H 'Content-Type':'application/json' \
    -d '{"username": "bart", "password": "mysecret"}' \
    http://treqs:changeit@localhost:8080/treqs2/administration/authentication/user/
```

By default, users are created with the role 'USER'

To create a new user 'homer' with password 'topsecret' and role 'ADMIN': 

```
> curl -X POST -H 'Content-Type':'application/json' \
    -d '{"username": "homer", "password": "topsecret", "role": "ADMIN"}' \
    http://treqs:changeit@localhost:8080/treqs2/administration/authentication/user/
```


To set role ADMIN to user 'lisa':

```
> curl -X PUT 'http://treqs:changeit@localhost:8080/treqs2/administration/authentication/user/lisa/role/ADMIN'
```

To deny access to resources for user 'burns':

```
> curl -X DELETE 'http://treqs:changeit@localhost:8080/treqs2/administration/authentication/user/burns/role/NONE'
```

You can also do this with the TReqS client.


## Managing the HPSS connection 

You can test TReqS without connecting it to a genuine HPSS system by using the 
embedded simulator called SimulatedHPSS.

### SimulatedHPSS

SimulatedHPSS provides metadata of files defined in 'FILES4TESTS/' directory and simulated staging using sleep() 
taking into account file size and based on reading rate of 200MiB/s.

To use it, provide SimulatedHPSS as HPSS connector in 'etc/treqs2.properties':

```
treqs.hpss.connector=fr.in2p3.cc.storage.treqs2.hsm.hpss.SimulatedHPSS
```

### Real HPSS instance

To work with real HPSS instance, you must set HPSS connector in 'etc/treqs2.properties':

```
treqs.hpss.connector=fr.in2p3.cc.storage.treqs2.hsm.hpss.HPSS
```

You must also provide required libraries and credentials.

#### About libs

* HPSS libs

	To communicate with HPSS, TReqS does use the HPSS library installed locally. By default TReqS expects it to be 
	installed in '/opt/hpss/lib' directory. If this is not the case on your system, please adjust 'etc/my-wrapper.conf':

	```
	set.default.LD_LIBRARY_PATH=</path/to/your/hpss/lib>:../lib/
	```

* log4cxx libs

	TReqS is written in Java and addresses HPSS C API using JNI. To log activity of JNI bridge, we use log4cxx, 
	which requires the liblog4cxx.so library. Don't forget to install it.

* libtirpc (mandatory with HPSS 7.4.2)

	With HPSS 7.4.2, it's mandatory to force loading libtirpc before all other libs. 
	(otherwise staging will hang with any error message)
	
	This is done using LD_PRELOAD environment variable. Please adjust 'etc/my-wrapper.conf' to your unix config:
	
	```
  # Force load of library libtirpc (mandatory with HPSS 7.4.2)
  # doing this will force loading libtirpc before all other libs
  set.LD_PRELOAD=/lib64/libtirpc.so
	```

#### About credentials   

In TReqS, HPSS authentication is ‘UNIX’ authentication

Username and associated keytab can be provided in config file 

Adjust credential properties 'treqs.hpss.auth.*' in 'etc/treqs2.properties':

```
treqs.hpss.auth.user=unix_username
treqs.hpss.auth.keytab=/path/to/keytab 
```

Default values are  ‘root’ and ‘/var/hpss/etc/keytab.root‘




## Managing staging processes

### Customizing staging

Different properties are available in the 'etc/treqs2.properties' file for customizing the staging process. See properties like 'treqs.dynamic.stager.*'.

For TReqS to compute a realistic timeout value, it is recommended to adjust rates for tape models (see "Managing tape models").


### Controlling the scheduler   

Staging tasks are triggered at fixed intervals, configurable by the property 'treqs.dynamic.stager.event_scheduler_rate'
in 'etc/treqs2.properties'. The choice of this value is tricky and must not be too short to get benefits from aggregating requests over time.

It's possible to disable the scheduler and obviously to enable it again.
Keep in mind that enabling the scheduler will immediately fire a new staging phase
in addition to reactivate the scheduler at fixed intervals.

To disable the scheduler (and consequently future potential staging processes):

```
> curl -X DELETE 'http://treqs:changeit@localhost:8080/treqs2/administration/scheduler/enabled'
```

To enable the scheduler (and consequently future staging processes) and immediately fire a new staging phase: 

```
> curl -X PUT 'http://treqs:changeit@localhost:8080/treqs2/administration/scheduler/enabled'
```

You can also do this with the TReqS client.


## Persistence

All entities are persisted into an embedded H2 database.
By default H2 files are located in 'data/' directory.
You can customize the H2 database location using the 'treqs.db.root_dir' property in the 'etc/treqs2.properties' file.

As a TReqS administrator, you are responsible for backing up the database.

## Logging

All TReqS activity is recorded in files using logback and log4cxx (for JNI part). 
By default, logs are located in 'logs/' directory.

You can customize how/where logs are written in the 'etc/logback.xml' and 'etc/log4cxx.xml'. 
Changes are taken into account at run-time.

## Archiving solutions

By default archived requests are written in the log file with the following pattern:

```
<Date> RequestHandler INFO  f.i.c.s.t.a.LogArchiveModule [ARCHIVE] <JSON representation of the archived request>
```

If you prefer to have archives written in a separate CSV file, modify the 'etc/treqs2.properties' as follows:

```
treqs.archive.module=fr.in2p3.cc.storage.treqs2.archiving.CSVArchiveModule
```
and adjust 'treqs.csvarchive.*' properties to customize the CSV content (fields and separator).

In this case, the administrator of TReqS of responsible for CSV file management (for example with the logrotate tool)

Ended requests stay in the database until they are removed on a regular basis, typically at fixed intervals of time.

It is also possible to write your own archive module, see "Advanced configuration".

## Secure HTTP

As passwords are transmitted in HTTP, it is strongly recommended to configure the TReqS web server to start in secure mode. Change
the 'etc/treqs2.properties' file as follows:

```
treqs.server.ssl=true
```
and adapt SSL parameters 'javax.net.ssl.*' at the end of the file.

The keystore contains the key associated to the certificate of the host. The truststore contains
certificates of all CA (certification authority) that accepts client certificates (if provided).
The truststore can be the same file as the keystore.

To generate the keystore with PKCS12 host certificate into it:

```
> keytool -noprompt -importkeystore -srckeystore hostcert.p12 -srcstoretype pkcs12 \
        -srcstorepass p12password -destkeystore keystore.jks -deststorepass changeit
```

to generating the truststore with the CA certificate into it:

```
> keytool -keystore truststore.jks -import -file CAname.pem -alias CAname -noprompt -storepass changeit
```

## Other configuration files

In 'etc/' you will also find configuration for

* JSW : wrapper.conf + my-wrapper.conf
* JAAS : jaas.config
* TReqS server : tapemodels.properties (see "Managing tape models")
* LOGS : logback.xml + log4cxx.xml (see "Logging")

In 'web/' directory you will find configuration for web pages content (stylesheets, scripts and templates).

# Tips and tricks

## Startup actions 

* DB creation

  TReqS can be started from scratch. This means that the DB is created at startup if it does not already exists. 
  In other word, you can remove the 'data/' directory, it will be created with a new fresh DB, cool !
  Moreover, the DB will be created with a default user (cf Users Management) and a default tape model (cf Tape Model management).

* DB reset
  
  Sometimes, for example during tests, it may be useful to clear the H2 DB in a smarter and lighter way than removing the 'data/' directory.
  
  This can be done at startup, by setting the environment variable
  TREQS_CLEAN_DB to the keyword 'yes_I_do_clean_DB'.
  Doing this, requests, files and queues tables will be cleaned, but users and tape models will remain unchanged. Interesting, isn't it ?
  
  ```
  > TREQS_CLEAN_DB="yes_I_do_clean_DB" bin/treqsd start (or restart)
  ```
  Caution : Do NOT set TREQS_CLEAN_DB in a permanent way ! 
  

 
# Advanced configuration

## Authentication  

TReqS is designed to accept any JAAS module. If you know how to use your own JAAS module, have a look at the JAAS configuration
file 'etc/jaas.config'. Modifying the JAAS configuration requires a server restart.

## Plug your own archive module

Please contact us to know more about developing and plugging your own archive module.

## Use an alternative JMS server

By default, TReqS will try to connect to an OpenMQ server running on the localhost on port 7676. As an alternative, it is possible
to run Apache ActiveMQ as JMS server.

### Configuration to run with Apache ActiveMQ 5.14

First change the name of the ConnectionFactory class name in 'etc/treqs2.properties':
  
  ```
  # JMS provider
  # Default provider is OpenMQ (from GlassFish) https://mq.java.net
  #treqs.jms.connectionfactory=com.sun.messaging.ConnectionFactory
  # But let's use ActiveMQ instead
  treqs.jms.connectionfactory=org.apache.activemq.ActiveMQConnectionFactory
  ```
Since 5.14, ActiveMQ has an enhanced security feature that requires to declare serializable packages. TReqS classes need to be added to
allowed serializable packages by setting the system property in 'etc/wrapper.conf':

  ```
  wrapper.java.additional.<NUMBER>=-Dorg.apache.activemq.SERIALIZABLE_PACKAGES=java.lang,javax.security,java.util,fr.in2p3.cc.storage.treqs2.core.entity
  ```

